/*
** Module   : ExpressionConversionWorker.java
** Abstract : provides a simple, callable interface to convert entire
**            expression trees from Progress to Java AST forms
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050406   @20862 Created initial version which supports a
**                           manual tree walking approach such that an
**                           expression can be converted "in-line" from
**                           any rule and the result will be properly
**                           located in the target tree.  This is critical
**                           since expressions are used *everywhere*, so
**                           the conversion is best done by calling this
**                           worker's high level routines when a node
**                           of type EXPRESSION is encountered.
** 002 GES 20050501   @20960 Added support for variable definitions and
**                           references.
** 003 GES 20050505   @21121 Moved the majority of expression processing
**                           into rule-sets and this leaves only the
**                           minimum helper methods here.  Some methods
**                           were moved here from other classes to
**                           centralize the few helpers that are needed
**                           specifically for expressions.
** 004 ECF 20050526   @21290 Changes to support new expression engine
**                           implementation. Changed library registration
**                           mechanism based on superclass' modifications
**                           to support single library per pattern worker
**                           limit.
** 005 GES 20050530   @21369 Migrated to new data type wrappers.
** 006 GES 20050603   @21399 Removed variable definition/init helpers
**                           because the worker is deleted and the logic
**                           exists in rule sets directly.
** 007 GES 20050607   @21445 Added a helper to determine the Java wrapper
**                           type associated with an expression.
** 008 GES 20050614   @21487 Updated FUNC_POLY processing to support the
**                           IF/THEN/ELSE function and to have better
**                           error handling.
** 009 GES 20050901   @22433 Removed operator processing and updated
**                           expressionType() to support the new method
**                           and attribute data types.
** 010 GES 20050928   @22873 Removed createDateReference() and simplified
**                           naming.
** 011 GES 20051024   @23111 Added conversion rule-set access to the
**                           convertToRegEx and convertToSQLLike methods
**                           of the character wrapper.
** 012 ECF 20051108   @23278 Added support for RECID/ROWID types in
**                           expressions. Both convert to integer in our
**                           implementation.
** 013 ECF 20060114   @23912 Added support for ACCUM function to
**                           expressionType. Return type depends on type
**                           of accumulation as well as expression being
**                           accumulated.
** 014 GES 20060228   @24788 Added support for INPUT function.
** 015 GES 20060302   @24846 Added support for RAW/MEMPTR types.
** 016 GES 20060308   @24933 Added support for CUSTOMER_SPECIFIC types.
** 017 ECF 20060321   @25170 Added compareClientWhereNodes user function.
**                           Helper function for client-side where clause
**                           processing.
** 018 GES 20060407   @25417 Fix for INPUT morphing mode.
** 019 GES 20060410   @25455 Better handling for unknown value literals
**                           in the ternary (IF) builtin function.
** 020 GES 20060421   @25638 Added a safe way to test if a given node is
**                           a valid part of an expression.
** 021 GES 20060421   @25642 Fix for infinite loop created when an INPUT
**                           function is compared to another INPUT 
**                           function.
** 022 GES 20060911   @29487 Exposed new helpers for string literal and
**                           format string processing. 
** 023 GES 20060920   @29720 Added formattedResultSize().
** 024 GES 20061018   @30481 Added minimal support for a symbol node in
**                           expressionType(). The only case supported is
**                           an in-line var def in a UI language stmt.
** 025 GES 20070907   @35014 Support for 10.1B data types.
** 026 GES 20080327   @37697 Support for handle types (how did I miss that
**                           before?).
** 027 ECF 20080730   @39284 Added first class support for rowid type.
**                           This type now has its own wrapper class. The
**                           integer wrapper is still used for recid.
** 028 GES 20090424   @41929 Import change.
** 029 SVL 20090225   @41373 Record IDs are mapped to recid, not to integer.
** 030 GES 20090518   @42379 Import change.
** 031 SIY 20090915   @43902 Added isFormatCompatible() method.
** 032 SVL 20110622          Added convertToSQLBegins,
**                           postprocessStringLiteral, toUpperCase.
** 033 SIY 20111101          Updated imports.
** 034 OM  20121218          Added support for int64 datatype.
** 035 EVL 20130107          Changing internal progressToJavaString()
**                           implementation to take into account possible
**                           Windows-target style of the '\' character in
**                           path related string handling.
** 036 CA  20130123          Provided support for DYNAMIC-FUNCTION.
** 037 OM  20130128          Added support for datetime(tz) data types.
** 038 CA  20130128          Updated reference from character to TextOps.
**                           Allow longchar expressions in concatenation.
** 039 OM  20130212          Added support for CONTAINS clause.
** 040 GES 20130225          Major improvement/cleanup to *_POLY node processing.
**                           Some new _POLY types are added and context analysis was
**                           added to the ones that can't be otherwise predicted.
** 041 GES 20130301          Removed some incorrect analysis in PLUS typing. Added a
**                           best efforts try to determine the MINUS case when the
**                           1st operand is missing.
** 042 GES 20130302          Fixed a ClassCastException.
** 043 GES 20130304          Added DB_REF_NON_STATIC support.
*/

package com.goldencode.p2j.convert;
                           
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.cfg.*;

/**
 * A pattern worker whose purpose is to provide helpers for conversion of
 * Progress language expression trees into the corresponding Java trees.  The
 * general idea is to call a user function (from a pattern engine rule set)
 * referencing a Progress source AST node that is part of a subtree that 
 * is rooted at type <code>EXPRESSION</code>.
 * <p>
 * In all expression conversion, the structure of the expression is
 * maintained, such that the generated target AST node(s) will be
 * parented/rooted at the corresponding 'peer' nodes in the source AST.
 * <p>
 * This class provides access to a static method in <code>character</code>
 * for string conversion, {@link ExpressionHelper#progressToJavaString}.
 * <p>
 * More importantly, this class provides an analysis method which is used
 * to determine the type of an expression or any sub-node of an expression.
 */
public final class ExpressionConversionWorker
extends AbstractPatternWorker
implements ProgressParserTokenTypes
{
   /** Nodes for which context analysis is active inside a *_POLY node. */
   private static Set<Long> recursionIds = new HashSet<Long>();
   
   /**
    * Default constructor which initializes libraries.
    */
   public ExpressionConversionWorker()
   {
      super();
      setLibrary(new ExpressionHelper());
   }

   /**
    * Analyzes the type of operands and infers the expression type for binary
    * operations (plus, minus, multiply, modulus), maybe others.
    * @param   ftype
    *          The type of the first operand.
    * @param   stype
    *          The type of the second operand.
    * 
    * @return  the type of the expression as string ("integer", "decimal", 
    *          "int64") if operand types are valid or null otherwise.
    */
   private static String getExpressionType(final String ftype,
                                           final String stype)
   {
      boolean bInt32 = false;
      boolean bInt64 = false;
      boolean bFloat = false;
      boolean bUnknown = false;

      if ("integer".equals(ftype))
      {
         bInt32 = true;
      }
      else if ("int64".equals(ftype))
      {
         bInt64 = true;
      }
      else if ("decimal".equals(ftype))
      {
         bFloat = true;
      }
      else
      {
         bUnknown = true;
      }

      if ("integer".equals(stype))
      {
         bInt32 = true;
      }
      else if ("int64".equals(stype))
      {
         bInt64 = true;
      }
      else if ("decimal".equals(stype))
      {
         bFloat = true;
      }
      else
      {
         bUnknown = true;
      }

      if (bUnknown)
      {
         // at least one operand is not numeric
         return null;
      }
      else if (bFloat)
      {
         // at least one of the operands is decimal
         return "decimal";
      }
      else if (bInt64)
      {
         // at least one of the operands is a long
         return "int64";
      }
      else
      {
         // both operands are 32 bit integers
         return "integer";
      }
   }

   /**
    * Walks the <code>ProgressAst</code> node specified and determines the
    * data type of the wrapper object returned by that sub-expression.  This 
    * method calls itself recursively as needed to determine the proper type.
    * <p>
    * At this time there is support for all basic data types (date, character,
    * integer, decimal, logical, recid, rowid, raw and memptr). The special
    * <code>FUNC_POLY, ATTR_POLY, METH_POLY</code> types are only supported for
    * the following:
    * <pre>
    * ABSOLUTE based on the operand
    * ACCUM based on the accumulator subtype and sometimes the expression operand
    * BUFFER-VALUE with context analysis
    * DYNAMIC-FUNCTION, with "casttype" annotation or content analysis
    * GET-BYTES
    * IF based on the 2nd (and sometimes 3rd) operand
    * INPUT the referent type and sometimes overridden with context analysis
    * MAXIMUM based on the list of parameters
    * MINIMUM based on the list of parameters
    * SUPER based on the return type of the containing function definition
    * </pre>
    * <p>
    * The INPUT built-in function normally returns the same type as the
    * single parameter (which must be an lvalue which is found in a frame
    * that is in scope).  However, this function can be used in an expression
    * that requires a <code>character</code> data type even in the case where
    * the parameter is NOT of the <code>character</code> type!  This is like
    * an override option and it must be detected from the SURROUNDING 
    * expression (above the FUNC_POLY node). This support is implemented 
    * except for one case: <b>the usage of INPUT as a parameter to another
    * built-in function (or presumably a method) which expects a 
    * <code>character</code> parameter will also cause this conversion. At
    * this time there is no support in these cases. The resulting type (if
    * not already <code>character</code>) will cause the wrong type to be
    * emitted.</b>
    * <p>
    * Attributes and methods are fully supported except for some
    * <code>ATTR_POLY and METH_POLY</code> nodes that are not processed.
    *
    * @param    source
    *           The AST node to be inspected.
    *
    * @return   The classname of the wrapper type returned by this 
    *           sub-expression.
    */
   public static String expressionType(Aast source)
   throws IllegalArgumentException,
          UnsupportedOperationException
   {
      if (source == null)
         throw new IllegalArgumentException("Source node cannot be null!");
      
      String  ftype    = null;
      String  stype    = null;
      String  errmsg   = null;
      int     type     = source.getType();
      int     oldtype  = -1;

      // first pass (may be null)
      String jcls = simpleExpressionType(type);
      
      // some of the simple cases may need additional processing and this also
      // handles the nodes that cannot be determined by type alone 
      switch (type)
      {
         case CUSTOMER_SPECIFIC:
            jcls = (String) source.getAnnotation("etype");
            break;
            
         case SYMBOL:
            // ONLY in-line var definitions (in UI stmts using a format
            // phrase) are supported at this time!
            Aast def = (Aast) source.getNextSibling();
            
            while (def != null && def.getType() != AGGREGATE)
            {
               def = (Aast) def.getNextSibling();
            }
            
            if (def != null && def.getType() == FORMAT_PHRASE)
            {
               if (def.isAnnotation("classname") &&
                   def.isAnnotation("vardef"))
               {
                  jcls = (String) def.getAnnotation("classname");
               }
            }
            break;
            
         case EXPRESSION:
         case LPARENS:
         case UN_MINUS:
         case UN_PLUS:
            jcls  = expressionType(source.getChildAt(0));
            break;
            
         case COLON:
            jcls  = expressionType(source.getChildAt(1));
            break;
            
         case MULTIPLY:
            // must be numeric but we need to know if it promotes to int64 / decimal
            ftype = expressionType(source.getChildAt(0));
            stype = expressionType(source.getChildAt(1));

            jcls = getExpressionType(ftype, stype);
            if (jcls == null)
            {
               // at least one operand is not numeric
               errmsg = "MULTIPLY cannot have op1 (" + ftype +
                     ") and op2 (" + stype + ").";
            }
            break;
            
         case KW_MOD:
            // if at least one of the operands are int64 the result is int64 
            ftype = expressionType(source.getChildAt(0));
            stype = expressionType(source.getChildAt(1));

            if ("decimal".equals(ftype) ||
                "integer".equals(ftype) ||
                "int64".equals(ftype))
            {
               if ("integer".equals(stype) ||
                   "int64".equals(stype))
               {
                  if ("decimal".equals(ftype))
                  { 
                     // result is the same as the 2nd operand
                     jcls = stype;
                  }
                  else
                  {
                     // both operands are integer-like type
                     jcls = getExpressionType(ftype, stype);
                  }
               }
            }
            if (jcls == null)
            {
               // at least one operand is not numeric
               errmsg = "MOD cannot have op1 (" + ftype +
                     ") and op2 (" + stype + ").";
               jcls = null;
            }
            break;
            
         case FUNC_INT:
            // possible override of the simple type
            oldtype = unpackOldType(source);
            if (oldtype == KW_RANDOM)
            {
               // if at least one of the operands are int64 the result is int64
               ftype = expressionType(source.getChildAt(0));
               stype = expressionType(source.getChildAt(1));

               jcls = getExpressionType(ftype, stype);
            }
            break;
            
         case NUM_LITERAL:
         case VAR_INT:
         case FIELD_INT:
         case ATTR_INT:
         case METH_INT:
            // possible override of the simple type
            Boolean use64bit = (Boolean) source.getAnnotation("use64bit");
            if (use64bit != null &&  use64bit.booleanValue())
            {
               jcls = "int64";
            } 
            break;
            
         case PLUS:
            ftype = expressionType(source.getChildAt(0));
            stype = expressionType(source.getChildAt(1));
            
            if (ftype == null)
            {
               // in this case we need to look at the 2nd operand and determine
               // what is possible from that (not as much as we would like since
               // numeric 2nd operands can be used with the various date types as
               // well as with numerics in the 1st operand)
               if ("character".equals(stype) || "longtype".equals(stype))
               {
                  jcls = stype;
               }
            }
            else if ("date".equals(ftype)       ||
                     "datetime".equals(ftype)   || // Adds a number of millis
                     "datetimetz".equals(ftype) || // Adds a number of millis
                     "longchar".equals(ftype))
            {
               jcls = ftype;
            }
            else if ("character".equals(ftype))
            {
               // longtype in either operand results in longtype, for the result
               // to be character, both operands must be character; regardless, in
               // this case the result will be the same as the 2nd operand
               jcls = stype;
            }
            else if ("integer".equals(ftype) ||
                     "int64".equals(ftype) ||
                     "decimal".equals(ftype))
            {
               // must be numeric but we need to know if it promotes to 
               // decimal or int64
               jcls = getExpressionType(ftype, stype);
               if (jcls == null)
               {
                  // at least one operand is not numeric
                  errmsg = "PLUS cannot have op2 (" + stype + ").";
                  jcls = null;
               }
            }
            else
            {               
               errmsg = "PLUS cannot have op1 (" + ftype + ").";
            }
            break;
            
         case MINUS:
            ftype = expressionType(source.getChildAt(0));
            
            if (ftype == null)
            {
               // in this case we need to look at the 2nd operand and determine
               // what is possible from that (not as much as we would like since
               // numeric 2nd operands can be used with the various date types as
               // well as with numerics in the 1st operand)
               if ("date".equals(stype)     ||
                   "datetime".equals(stype) ||
                   "datetimetz".equals(stype))
               {
                  jcls = stype;
               }
            }
            else if ("date".equals(ftype))
            {
               // check the second child to guess which form operator is used
               stype = expressionType(source.getChildAt(1));
               jcls  = "date".equals(stype) ? "integer" : "date";
            }
            else if ("datetime".equals(ftype) || "datetimetz".equals(ftype))
            {
               // check the second child to guess which form operator is used
               stype = expressionType(source.getChildAt(1));
               jcls  = ("datetime".equals(stype) || 
                        "datetimetz".equals(stype)) ? 
                       "int64" :
                       stype;
            }
            else if ("integer".equals(ftype) ||
                     "decimal".equals(ftype) ||
                     "int64".equals(ftype))
            {
               // must be numeric but we need to know if it promotes to 
               // decimal or int64
               stype = expressionType(source.getChildAt(1));

               jcls = getExpressionType(ftype, stype);
               if (jcls == null)
               {
                  // at least one operand is not numeric
                  errmsg = "MINUS cannot have op2 (" + stype + ").";
                  jcls = null;
               }
            }
            else
            {
               errmsg = "MINUS cannot have op1 (" + ftype + ").";
            }
            break;
            
         // polymorphic return values need lots of help to determine the real type
         case DB_REF_NON_STATIC:
            // try detecting from context
            jcls = analyzeContext(source);
            
            // fall back to a "generic" result
            if (jcls == null)
            {
               jcls = "BaseDataType";
            }
            break;
            
         // polymorphic return values need lots of help to determine the real type
         case FUNC_POLY:
         case ATTR_POLY:
         case METH_POLY:
            
            oldtype = unpackOldType(source);
            Aast child = source.getChildAt(0);
            
            switch (oldtype)
            {
               case KW_ABS:
                  // return the same type as the first operand
                  jcls = expressionType(child);
                  break;
                  
               case KW_ACCUM:
                  // get the child node of AGGREGATE which describes the type
                  // of accumulation we're doing
                  Aast accumTypeAst = (Aast) child.getFirstChild();
                  int accumType = accumTypeAst.getType();
                  // get the expression AST, which is AGGREGATE's next
                  // sibling
                  Aast accumExprAst = (Aast) child.getNextSibling();
                  switch (accumType)
                  {
                     case KW_COUNT:
                     case KW_SUB_CNT:
                        jcls = "integer"; // TODO: OM: investigate int64
                        break;
                     case KW_AVERAGE:
                     case KW_SUB_AVG:
                        jcls = "decimal";
                        break;
                     case KW_MAX:
                     case KW_SUB_MAX:
                     case KW_MIN:
                     case KW_SUB_MIN:
                     case KW_TOTAL:
                     case KW_SUB_TOT:
                        jcls = expressionType(accumExprAst);
                        break;
                     default:
                        errmsg = "Unknown " + source.lookupTokenName(type) + 
                                 " (" + source.lookupTokenName(accumType) +
                                 ")";
                  }
                  break;
                  
               case KW_ADD_INVL:
                  // return the same type as the first operand which should always be
                  // one of date, datetime or datetimetz
                  jcls = expressionType(child);
                  break;
                  
               case KW_BUF_VAL:
                  // try detecting from context
                  jcls = analyzeContext(source);
                  
                  // complain if we have no other solution
                  if (jcls == null)
                  {
                     String msg1 = "WARNING: no type calculated for BUFFER-VALUE in [%d] %s";
                     System.out.println(String.format(msg1, source.getId(), source.dumpTree()));
                  }
                  break;
                  
               case KW_DYN_FUNC:
                  // check if the tree already has an annotation that tells us the type
                  jcls = (String) source.getAnnotation("casttype");
                  
                  if (jcls == null)
                  {
                     // try detecting from context
                     jcls = analyzeContext(source);
                  }
                  
                  // complain if we have no other solution
                  if (jcls == null)
                  {
                     String msg2 = "WARNING: 'casttype' return type annotation " +
                                  "expected for the DYNAMIC-FUNCTION call [%d] %s";
                     System.out.println(String.format(msg2, source.getId(), source.dumpTree()));
                  }
                  break;
                  
               case KW_GET_BYTS:
                  // TODO: determine what is really possible here, some sample code
                  //       has been seen where the result is assigned directly to an
                  //       integer like this: my-int-var = GET-BYTES(some-byte-source, my-pos, 2)
                  //       Perhaps the value is always returned as a memptr and then
                  //       the return value must naturally be handled by the context.
                  //       Or perhaps this is wrong.
                  jcls = "memptr";
                  break;
                  
               case KW_IF:
                  // by default return the same type as the second operand
                  jcls = expressionType(source.getChildAt(1));

                  boolean utype = "unknown".equals(jcls);
                  
                  // special cases:
                  // 1. integers and decimals are interchangeable so if the
                  //    next operand is a decimal, then our return type is
                  //    really decimal too
                  // 2. unknown value is not a valid type in and of itself,
                  //    there must be a real type in the 3rd operand which
                  //    can be used as the type for both operands
                  if ("int64".equals(jcls) || "integer".equals(jcls) || utype)
                  {
                     String jcls2 = expressionType(source.getChildAt(2));
                     
                     // unknown op1 or (integer op1 + non-unknown op2)
                     if (utype || !"unknown".equals(jcls2)) 
                     {
                        // the net of this is that in this case the return
                        // type will always be the same as the third operand
                        jcls = jcls2;
                     }
                  }
                  
                  break;
               
               case KW_INPUT:
                  // find the real child parameter
                  if (child.getType() == KW_FRAME ||
                      child.getType() == KW_BROWSE)
                  {
                     child = source.getChildAt(1);
                  }
               
                  // normal return type is the same as the referent
                  jcls = expressionType(child);
                  
                  // INPUT has a special mode that can return a character
                  // type if the SURROUNDING expression requires it (BUT 
                  // WE DON'T ENTER THIS BLOCK if there is another INPUT
                  // function being analyzed already because we must avoid
                  // infinite loops)
                  if (!jcls.equals("character"))
                  {
                     String otype = analyzeContext(source);

                     if (otype != null && otype.equals("character"))
                     {
                        jcls = "character";
                     }
                  }
                  break;
                  
               case KW_MAX:
               case KW_MIN:
                  boolean had64 = false;
                  
                  // these functions return a type based on the list of operands
                  while (child != null)
                  {
                     jcls = expressionType(child);
                     
                     if ("int64".equals(jcls))
                     {
                        had64 = true;
                     }
                     
                     // TODO: check that the following can't be used in MIN/MAX:
                     //       class, com-handle, handle, memptr, raw, recid,
                     //       rowid, blob (field), clob (field) 
                     
                     // if we are integer or int64 we allow the loop to continue,
                     // and if all items are the same then this will be returned
                     // at the end OR otherwise it will be "promoted" to
                     // decimal
                     if ("character".equals(jcls)  ||
                         "longchar".equals(jcls)   ||
                         "logical".equals(jcls)    ||
                         "date".equals(jcls)       ||
                         "datetime".equals(jcls)   ||
                         "datetimetz".equals(jcls) ||
                         "decimal".equals(jcls))
                     {
                        // we have found the type so we exit our loop
                        break;
                     }
                     else if ("unknown".equals(jcls))
                     {
                        errmsg = "Unknown value cannot be a parameter for " +
                                 source.lookupTokenName(oldtype);
                        break;
                     }  
                     
                     child = (Aast) child.getNextSibling();
                  }
                  
                  // widen to int64 if all values are only ever integer/int64 
                  if ("integer".equals(jcls) && had64)
                  {
                     jcls = "int64";
                  }
                  
                  if (jcls == null)
                  {
                     errmsg = "Invalid function call - no parameters for " +
                              source.lookupTokenName(oldtype);
                  }
                  break;
                  
               case KW_SUPER:
                  // based on the return type of the containing function definition
                  Aast function = source.getAncestor(-1, FUNCTION);
                  Aast funcdef  = function.getChildAt(0);
                  long functype = (Long) funcdef.getAnnotation("type");
                  jcls = simpleExpressionType((int) functype);
                  break;
                  
               // unexpected *_POLY node, error out here
               default:
                  errmsg = "Unknown " + source.lookupTokenName(type) + 
                           " with oldtype of " + source.lookupTokenName(oldtype);
            }
            break;
            
         default:
            if (jcls == null)
            {
               errmsg = "Unknown node type " + source.lookupTokenName(type);
            }
      }
      
      if (errmsg != null)
      {
         throw new UnsupportedOperationException(errmsg);
      }
      
      return jcls;
   }
   
   /**
    * Return the value of the given node's <code>oldtype</code> annotation.
    *
    * @param    node
    *           The node from which to obtain the annotation.
    *
    * @return   The oldtype or -1 if there is no such annotation.
    */
   public static int unpackOldType(Aast node)
   {
      int otype = -1;
      
      if (node.isAnnotation("oldtype"))
      {
         Long oldtype = (Long) node.getAnnotation("oldtype");
         otype = oldtype.intValue(); 
      }
      
      return otype;
   }
   
   /**
    * Attempt to detect the type of the given node by looking at the context
    * above and to either side, instead of down.
    * <p>
    * At this time, this has simplistic recursion prevention.  On entry, the node
    * ID of the <code>source</code> is added to a set. If that node is already in
    * the set, it forces an immediate return of <code>null</code>. This protects
    * from possible infinite loops.
    *
    * @param    source
    *           The node to analyze.
    *
    * @return   The calculated wrapper type or <code>null</code> if no type could
    *           be determined.
    */
   public static String analyzeContext(Aast source)
   {
      long currentId = source.getId();
      
      String otype = null;
      String ftype = null;
      
      if (!recursionIds.contains(currentId))
      {
         try
         {
            // add our node ID to the flag
            recursionIds.add(currentId);
            
            Aast   past  = source.getParent();
            int    ptype = past.getType();
            int    index = source.getIndexPos();
            
            // quick out for ATTR_POLY that is index 0 with a COLON parent, this
            // by definition must be a handle
            if (source.getType() == ATTR_POLY && ptype == COLON && index == 0)
            {
               return "handle";
            }
            
            // other ATTR_POLY cases must be checked below, but we must walk up
            // to get above the COLON ancestors; any chaining that isn't with us
            // as a handle (see above) must be the terminal node of the chain, so
            // we only need to "look up" the chain
            
            while (ptype == LPARENS || ptype == COLON || ptype == DB_REF_NON_STATIC)
            {
               // note the order of operations here
               index = past.getIndexPos();
               past  = past.getParent();
               ptype = past.getType();
            }
            
            int pold  = unpackOldType(past);
            
            switch (ptype)
            {
               // TODO: these datatype conversion funcs are only right for
               //       the 1st parameter... otherwise this generates the
               //       wrong answer
               case FUNC_CHAR:
                  if (pold == KW_STRING || pold == KW_CHR ||
                      pold == KW_CP_CVT || pold == KW_HEX_ENCD)
                  {
                     otype = "character";
                  }
                  break;
               case FUNC_DEC:
                  if (pold == KW_DEC)
                  {
                     otype = "decimal";
                  }
                  break;
               case FUNC_INT:
                  if (pold == KW_INT || pold == KW_ASC)
                  {
                     otype = "integer";
                  }
                  break;
               case FUNC_INT64:
                  if (pold == KW_INT64)
                  {
                     otype = "int64";
                  }
                  break;
               case FUNC_DATE:
                  if (pold == KW_DATE)
                  {
                     otype = "date";
                  }
                  break;
               case FUNC_DATETIME:
                  if (pold == KW_DATETIME)
                  {
                     otype = "datetime";
                  }
                  break;
               case FUNC_DATETIME_TZ:
                  if (pold == KW_DATE_TZ)
                  {
                     otype = "datetimetz";
                  }
                  break;
               case FUNC_RAW:
                  if (pold == KW_HEX_DECD)
                  {
                     otype = "raw";
                  }
                  break;
               case FUNC_MEMPTR:
                  if (pold == KW_BASE64_D)
                  {
                     otype = "memptr";
                  }
                  break;
               case FUNC_LONGCHAR:
                  if (pold == KW_BASE64_E)
                  {
                     otype = "longchar";
                  }
                  break;
               case FUNC_CLASS:
                  if (pold == KW_CAST || pold == KW_DYN_CAST)
                  {
                     // TODO: probably should read the proper classname from an
                     //       annotation
                     otype = "class";
                  }
                  break;
                  
               // ternary expression
               case FUNC_POLY:
                  if (pold == KW_IF)
                  {
                     if (index == 0)
                     {
                        // this is the condition expression
                        otype = "logical";
                     }
                     else
                     {
                        // based our type on the type of the other operand (not the
                        // condition)
                        otype = expressionType(past.getChildAt((index == 1) ? 2 : 1));
                     }
                  }
                  break;
                  
               case UN_MINUS:   
               case UN_PLUS:
               case MULTIPLY:
               case DIVIDE:
               case KW_MOD:
                  // TODO: we have picked the widest type here, but it may not be
                  //       correct (any numeric could be used)
                  otype = "decimal";
                  break;
                  
               case PLUS:
                  // not all cases can be detected
                  ftype = expressionType(past.getChildAt((index == 0) ? 1 : 0));
                  
                  if ("date".equals(ftype)     ||
                      "datetime".equals(ftype) ||
                      "datetimetz".equals(ftype))
                  {
                     // we must be index == 1 in this case
                     
                     // TODO: we have picked the widest type here, but it may not be
                     //       correct (any numeric could be used)
                     otype = "decimal";
                  }
                  else if ("character".equals(ftype) || "longchar".equals(ftype))
                  {
                     // TODO: does not "widen" properly
                     otype = ftype;
                  }
                  else if (index == 1 &&
                           ("integer".equals(ftype) ||
                            "int64".equals(ftype)   ||
                            "decimal".equals(ftype)))
                  {
                     // TODO: we have picked the widest type here, but it may not be
                     //       correct (any numeric could be used)
                     otype = "decimal";
                  }
                  break;
                  
               case MINUS:
                  // not all cases can be detected
                  ftype = expressionType(past.getChildAt((index == 0) ? 1 : 0));
                  
                  if (index == 0 && 
                      ("date".equals(ftype)     ||
                       "datetime".equals(ftype) ||
                       "datetimetz".equals(ftype)))
                  {
                     // TODO: this may not be quite right because different types
                     //       can be intermixed
                     otype = ftype;
                  }
                  else if (index == 1 &&
                           ("integer".equals(ftype) ||
                            "decimal".equals(ftype) ||
                            "int64".equals(ftype)))
                  {
                     // TODO: we have picked the widest type here, but it may not be
                     //       correct (any numeric could be used)
                     otype = "decimal";
                  }
                  break;
                  
               case GT:
               case LT:
               case GTE:
               case LTE:
               case EQUALS:
               case NOT_EQ:
                  otype = expressionType(past.getChildAt((index == 0) ? 1 : 0));
                  break;
                  
               case KW_NOT:
               case KW_AND:
               case KW_OR:
                  otype = "logical";
                  break;
                  
               case KW_BEGINS:
               case KW_MATCHES:
               case KW_CONTAINS:
                  otype = "character";
                  break;
                  
               case ASSIGN:
                  // we must be an lvalue, return the rvalue type
                  if (index != 0)
                  {
                     String msg1 = "WARNING: expected an lvalue and index != 0 [%d] %s";
                     System.out.println(String.format(msg1, source.getId(), source.dumpTree()));
                  }
                  otype = expressionType(past.getChildAt(1));
                  break;
                  
               case EXPRESSION:
                  Aast grampa = past.getParent();
                  int ttype = grampa.getType();
                  
                  switch (ttype)
                  {
                     case ASSIGN:
                        // if this is an assignment, check the type of the lvalue
                        otype = expressionType(grampa.getChildAt(0));
                        break;
                     case KW_WHEN:
                        otype = "logical";
                        break;
                     case KW_IF:
                        if (grampa.getParent().getType() == STATEMENT)
                        {
                           otype = "logical";
                        }
                        break;
                  }
                  break;
            }
         }
         
         finally
         {
            // clear the node ID from the set flag
            recursionIds.remove(currentId);
         }
      }
      
      return otype;
   }
   
   /**
    * Convert the type directly to a wrapper classname, if possible.  Since there
    * is no tree provided, a <code>null</code> is returned if no simple conversion
    * is possible.
    *
    * @param    type
    *           The type to convert.
    *
    * @return   The classname of the wrapper type for this type or <code>null</code> 
    *           if no simple conversion is possible.
    */
   public static String simpleExpressionType(int type)
   {
      String jcls = null;
      
      switch (type)
      {
         case DIVIDE:
         case DEC_LITERAL:
         case VAR_DEC:
         case FIELD_DEC:
         case FUNC_DEC:
         case ATTR_DEC:
         case METH_DEC:
            jcls  = "decimal";
            break;
            
         // these may need extra processing!
         case NUM_LITERAL:
         case VAR_INT:
         case FIELD_INT:
         case FUNC_INT:
         case ATTR_INT:
         case METH_INT:
            jcls = "integer";
            break;
            
         case VAR_RECID:
         case FIELD_RECID:
         case FUNC_RECID:
            jcls = "recid";
            break;
            
         case VAR_ROWID:
         case FIELD_ROWID:
         case FUNC_ROWID:
            jcls = "rowid";
            break;
            
         case VAR_INT64:
         case FIELD_INT64:
         case FUNC_INT64:
         case ATTR_INT64:
         case METH_INT64:
            jcls = "int64";
            break;
            
         case VAR_CLASS:
         case FIELD_CLASS:
         case FUNC_CLASS:
         case ATTR_CLASS:
         case METH_CLASS:
            // TODO: do we need to read an annotation to get the actual
            //       type of the object?
            jcls  = "class";
            break;
            
         case KW_BEGINS:
         case KW_MATCHES:
         case KW_CONTAINS:
         case GT:
         case LT:
         case GTE:
         case LTE:
         case EQUALS:
         case NOT_EQ:
         case KW_NOT:
         case KW_AND:
         case KW_OR:
         case BOOL_TRUE:
         case BOOL_FALSE:
         case FUNC_LOGICAL:
         case VAR_LOGICAL:
         case FIELD_LOGICAL:
         case ATTR_LOGICAL:
         case METH_LOGICAL:
            jcls = "logical";
            break;
            
         case STRING:
         case VAR_CHAR:
         case FIELD_CHAR:
         case FUNC_CHAR:
         case ATTR_CHAR:
         case METH_CHAR:
            jcls = "character";
            break;
            
         case VAR_LONGCHAR:
         case FUNC_LONGCHAR:
         case ATTR_LONGCHAR:
         case METH_LONGCHAR:
            jcls = "longchar";
            break;
            
         case FIELD_BLOB:
            jcls = "blob";
            break;
            
         case FIELD_CLOB:
            jcls = "clob";
            break;
            
         case DATE_LITERAL:
         case VAR_DATE:
         case FIELD_DATE:
         case FUNC_DATE:
         case ATTR_DATE:
         case METH_DATE:
            jcls = "date";
            break;
            
         case VAR_DATETIME:
         case FIELD_DATETIME:
         case FUNC_DATETIME:
         case ATTR_DATETIME:
         case METH_DATETIME:
            jcls = "datetime";
            break;
            
         case VAR_DATETIME_TZ:
         case FIELD_DATETIME_TZ:
         case FUNC_DATETIME_TZ:
         case ATTR_DATETIME_TZ:
         case METH_DATETIME_TZ:
            jcls = "datetimetz";
            break;
            
         case VAR_HANDLE:
         case FIELD_HANDLE:
         case FUNC_HANDLE:
         case ATTR_HANDLE:
         case METH_HANDLE:
            jcls  = "handle";
            break;
            
         case VAR_RAW:
         case FIELD_RAW:
         case FUNC_RAW:
         case ATTR_RAW:
         case METH_RAW:
            jcls  = "raw";
            break;
            
         case VAR_MEMPTR:
         case FUNC_MEMPTR:
         case ATTR_MEMPTR:
         case METH_MEMPTR:
            jcls  = "memptr";
            break;
            
         case UNKNOWN_VAL:
            jcls = "unknown";
            break;
            
         case METH_VOID:
            jcls = "void";
            break;
      }
      
      return jcls;
   }

   /**
    * Provides the core user-functions exported to pattern engine rules for
    * Progress to Java expression conversion.
    */
   public class ExpressionHelper
   {      
      /**
       * Exposes a helper method to convert Progress string literals into
       * the correct Java string literal using the static method
       * <code>character.progressToJavaString()</code>.
       *
       * @param    text
       *           A valid Progress string literal enclosed in double or 
       *           single quotes and possibly with string options.
       *
       * @return   The corresponding Java string literal.
       */
      public String progressToJavaString(String text)
      {
         boolean windows = !Configuration.getParameter("unix-escapes", true);
         return character.progressToJavaString(text, windows);
      }
      
      /**
       * Exposes a helper method to convert a Java string literal into
       * a runtime Java string (by processing the text of all escape
       * sequences into the corresponding character).
       *
       * @param    text
       *           A valid Java string literal (not enclosed in double 
       *           quotes).
       *
       * @return   The corresponding Java runtime string.
       */
      public String processEscapes(String text)
      {
         return StringHelper.processEscapes(text);
      }
      
      /**
       * Expands Progress format strings into the full text, explicit form 
       * of the format string. For example, "x(8)" would be returned as
       * "xxxxxxxx".
       *
       * @param    fmt
       *           A valid Progress format string.
       *
       * @return   The expanded format string.
       */
      public String preprocessFormatString(String fmt)
      {
         return character.preprocessFormatString(fmt);
      }
      
      /**
       * Calculates the correct maximum size of the text output using the
       * given format string for the given data type.
       * <p>
       * The given format should not have had its escape sequences processed
       * into a runtime form.
       *
       * @param    node
       *           The expression or sub-expression node for which the format
       *           string is to be used.
       * @param    fmt
       *           A valid format string.
       * @param    java
       *           <code>false</code> if the given string is as encoded in
       *           the original source file (it should not have been
       *           converted to a Java string).  <code>true</code> if this
       *           format string is text that is valid for encoding in a 
       *           Java source file.
       *
       * @return   The maximum number of characters that can be generated
       *           using this format.
       */
      public int formattedResultSize(Aast node, String fmt, boolean java)
      {
         if (!java)
            fmt = progressToJavaString(fmt);
         
         fmt = processEscapes(fmt);
         fmt = preprocessFormatString(fmt);
         
         return ("logical".equals(expressionType(node)))
                ? logical.maximumStringSize(fmt)
                : fmt.length();
      }
      
      /**
       * Exposes a helper method to convert Progress match patterns into
       * regular expressions using the static method
       * <code>character.convertToRegEx()</code>.
       *
       * @param    text
       *           A valid Progress match pattern.
       *
       * @return   The corresponding regular expression.
       */
      public String convertToRegEx(String text)
      {
         return character.convertToRegEx(text);
      }
      
      /**
       * Exposes a helper method to convert Progress match patterns into
       * SQL LIKE match expressions using the static method
       * <code>character.convertToSQLLike()</code>.
       *
       * @param    text
       *           A valid Progress match pattern.
       *
       * @return   The corresponding SQL LIKE match expression.
       */
      public String convertToSQLLike(String text)
      {
         return character.convertToSQLLike(text, false);
      }

      /**
       * Exposes a helper method to convert a second operand of the "op1
       * BEGINS op2" function into corresponding SQL LIKE expression using the
       * static method
       * {@link com.goldencode.p2j.util.character#convertToSQLBegins(String, boolean)}
       *
       * @param    text
       *           A second operand of the "op1 BEGINS op2" function.
       *
       * @return   The corresponding SQL LIKE expression.
       */
      public String convertToSQLBegins(String text)
      {
         return character.convertToSQLBegins(text, false);
      }

      /**
       * Exposes a helper method to convert a second operand of the "op1
       * CONTAINS op2" function into corresponding SQL LIKE expression using the
       * static method
       * {@link com.goldencode.p2j.util.character#convertToSQLContains(String, boolean)}
       *
       * @param    text
       *           A second operand of the "op1 BEGINS op2" function.
       *
       * @return   The corresponding SQL LIKE expression.
       */
      public String convertToSQLContains(String text)
      {
         return character.convertToSQLContains(text, false);
      }

      /**
       * Exposes a helper method which allows to return the uppercased
       * version of the runtime or hardcoded string using the static method
       * {@link com.goldencode.p2j.util.TextOps#toUpperCase(String, boolean)}
       *
       * @param    text
       *           The string to convert
       * @param    runtimeString
       *           <code>true</code> if the target string is a runtime string,
       *           <code>false</code> if it is a hardcoded string into a
       *           source file.
       *
       * @return   The uppercased string instance.
       */
      public String toUpperCase(String text, boolean runtimeString)
      {
         return TextOps.toUpperCase(text, runtimeString);
      }

      /**
       * Exposes a helper method which allows to postprocess string literal
       * using the static method
       * {@link com.goldencode.p2j.util.character#postprocessStringLiteral(String, boolean)}
       *
       * @param    text
       *           The string literal to process (without enclosing quotes)
       *
       * @return   processed string literal.
       */
      public String postprocessStringLiteral(String text)
      {
         return character.postprocessStringLiteral(text, false);
      }

      /**
       * Walks the <code>ProgressAst</code> node specified and determines the
       * data type of the wrapper object returned by that sub-expression.  This 
       * method processes the subtree recursively as needed to determine the 
       * proper type.
       * <p>
       * At this time there is support for all basic data types (date,
       * character, integer, decimal, logical, recid, rowid, raw and memptr).
       * The special <code>FUNC_POLY</code> type is only supported for the
       * following built-in functions:
       * <pre>
       * ABSOLUTE
       * IF
       * INPUT
       * MAXIMUM
       * MINIMUM
       * </pre>
       * <p>
       * The INPUT built-in function normally returns the same type as the
       * single parameter (which must be an lvalue which is found in a frame
       * that is in scope).  However, this function can be used in an
       * expression that requires a <code>character</code> data type even in
       * the case where the parameter is NOT of the <code>character</code>
       * type!  This is like an override option and it must be detected from 
       * the SURROUNDING expression (above the FUNC_POLY node). This support 
       * is implemented except for one case: <b>the usage of INPUT as a
       * parameter to another built-in function (or presumably a method) which 
       * expects a <code>character</code> parameter will also cause this
       * conversion. At this time there is no support in these cases. The
       * resulting type (if not already <code>character</code>) will cause the
       * wrong type to be emitted.</b>
       * <p>
       * Attributes and methods are fully supported except for the
       * <code>ATTR_POLY and METH_POLY</code> which are recognized but no
       * specific cases are currently handled.  In particular, the
       * <code>BUFFER-VALUE</code> is particularly difficult and has not been
       * resolved. 
       *
       * @param    source
       *           The AST node to be inspected.
       *
       * @return   The classname of the wrapper type returned by this 
       *           sub-expression.
       */
      public String expressionType(Aast source)
      {
         return ExpressionConversionWorker.expressionType(source);
      }
      
      /**
       * Check if specified format string is compatible with type of
       * expression passed as a parameter.
       * 
       * @param    source
       *           The expression to check compatibility with.
       * @param    format
       *           Format string.
       * 
       * @return   <code>true</code> if expression and format string are
       *           compatible with each other.
       */
      public boolean isFormatCompatible(Aast source, String format)
      {
         try
         {
            String type = ExpressionConversionWorker.expressionType(source);
            format = progressToJavaString(format);

            try
            {
               if ("character".equals(type))
               {
                  new StringFormat(format);
               }
               else if ("integer".equals(type) || "decimal".equals(type))
               {
                  new NumberFormat(format);
               } 
               else if ("date".equals(type))
               {
                  new DateFormat(format);
               }
               else if ("logical".equals(type))
               {
                  new BoolFormat(format);
               }
               
               return true;
            }
            catch (Exception e)
            {
               return false;
            }
         }
         catch (Exception e)
         {
            return false;
         }
      }
      
      /**
       * Walks the <code>ProgressAst</code> node specified and determines if
       * the given node is a valid part of an expression.
       *
       * @param    source
       *           The AST node to be inspected.
       *
       * @return   <code>true</code> if the given node is a valid expression
       *           or sub-expression node.  If <code>false</code>, then this
       *           node and its subtree cannot be processed as part of an
       *           expression.
       */
      public boolean isExpressionNode(Aast source)
      {
         String cls = null;
         
         try
         {
            cls = ExpressionConversionWorker.expressionType(source);
         }
         
         catch (UnsupportedOperationException uoe)
         {
            // let the cls be null
         }
         
         return (cls != null && cls.length() > 0);
      }
      
      /**
       * Helper user function to compare two Progress AST subtrees of a
       * database where clause expression, which must be evaluated on the
       * database client.  This is implemented as a pattern worker user
       * function, rather than a rules-based, named function, because the
       * latter does not support recursion, which this algorithm requires.
       * <p>
       * The two subtrees are considered equivalent if they have the same
       * structure, i.e, same token types arranged in the same locations.
       * In addition, each node must pass the following tests:
       * <ul>
       *   <li>for field node types, "schemaname" annotations must match;
       *   <li>for function node types, "oldtype" annotations must match;
       *   <li>for all other node types, text must match (case insensitive).
       * </ul>
       *
       * @param   node1
       *          Root of first subtree.
       * @param   node2
       *          Root of second subtree.
       *
       * @return  <code>true</code> if subtrees match, else
       *          <code>false</code>.
       */
      public boolean compareClientWhereNodes(Aast node1, Aast node2)
      {
         boolean same = false;
         int type1 = node1.getType();
         int type2 = node2.getType();
         
         if (node1 != null && node2 != null && type1 == type2)
         {
            // Compare field nodes.
            if (type1 > BEGIN_FIELDTYPES &&
                type1 < END_FIELDTYPES)
            {
               same = node1.getAnnotation("schemaname").equals(
                      node2.getAnnotation("schemaname"));
            }
            // Compare function nodes.
            else if (type1 > BEGIN_FUNCTYPES &&
                     type1 < END_FUNCTYPES)
            {
               same = node1.getAnnotation("oldtype").equals(
                      node2.getAnnotation("oldtype"));
            }
            // Compare all other nodes.
            else
            {
               same = node1.getText().equalsIgnoreCase(node2.getText());
            }
         }
         
         // Process children.
         if (same)
         {
            node1 = (Aast) node1.getFirstChild();
            node2 = (Aast) node2.getFirstChild();
            while (same && (node1 != null || node2 != null))
            {
               same = compareClientWhereNodes(node1, node2);
               node1 = (Aast) node1.getNextSibling();
               node2 = (Aast) node2.getNextSibling();
            }
         }
         
         return same;
      }
   }
}
